/*                                                                           *
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: amodigli $
 * $Date: 2011/05/09 06:34:48 $
 * $Revision: 1.1 $
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <xsh_efficiency_response.h>
#include <xsh_error.h>

HIGH_ABS_REGION NirHighAbsRegions[] = {
  {1060.0, 1232.9}, // nm
  {1247.4, 1500.0}, // nm
  {1549.0, 1588.5}, // nm
  {1595.0, 1625.0}, // nm
  {1630.0, 1983.5}, // nm
  {1984.2, 1988.1}, // nm
  {1988.9, 1991.1}, // nm
  {1991.7, 2138.0}, // nm
  {2145.0, 2191.0}, // nm
  {2218.0, 2271.0}, // nm
  {2351.0, 2431.0}, // nm
  {0., 0.} 
} ;

HIGH_ABS_REGION VisHighAbsRegions[] = {
  {585.0, 604.0}, // nm
  {684.5, 746.0}, // nm
  {755.5, 775.5}, // nm
  {783.7, 856.0}, // nm
  {887.5, 993.4}, // nm
  {0., 0.}
} ;



HIGH_ABS_REGION * 
xsh_fill_high_abs_regions(
   xsh_instrument* instrument,
   cpl_frame* high_abs_frame)
{
   HIGH_ABS_REGION * phigh=NULL;
   int nrow=0;
   double* pwmin=0;
   double* pwmax=0;
   int i=0;
   cpl_table* high_abs_tab=NULL;
   XSH_ARM the_arm;

  if(high_abs_frame !=NULL) {
     high_abs_tab=cpl_table_load(cpl_frame_get_filename(high_abs_frame),1,0);
  }
  the_arm=xsh_instrument_get_arm(instrument);



   if(high_abs_tab!=NULL) {
      nrow=cpl_table_get_nrow(high_abs_tab);
      check(pwmin=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MIN"));
      check(pwmax=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MAX"));

      phigh = (HIGH_ABS_REGION *) cpl_calloc(nrow + 1, sizeof(HIGH_ABS_REGION));
      for(i=0;i<nrow;i++) {
         phigh[i].lambda_min=pwmin[i];
         phigh[i].lambda_max=pwmax[i];
      }
      phigh[nrow].lambda_min=0;
      phigh[nrow].lambda_max=0;

   } else {

      /* Now specific to high absorption region (linear interpolation)
         NIR and VIS only
      */
      if ( the_arm == XSH_ARM_VIS ) {
         /* Specific VIS */
         phigh = VisHighAbsRegions ;
      }
      else if ( the_arm == XSH_ARM_NIR ) {
         /* Specific NIR */
         phigh = NirHighAbsRegions ;
      }

   }
  cleanup:
   xsh_free_table(&high_abs_tab);
   return phigh;
}
